@@ -13,10 +13,13 @@
 # limitations under the License.
 
 import copy
+import datetime
+import functools
 import logging
+import re
 import traceback
 from abc import abstractmethod
-from typing import List, Optional, Dict
+from typing import List, Optional
 
 from google.protobuf import text_format
 from prometheus_client import Histogram
@@ -26,43 +29,43 @@ from th2_grpc_common.common_pb2 import Event
 from th2_check2_recon.common import EventUtils, MessageComponent
 from th2_check2_recon.handler import MessageHandler, AbstractHandler
 from th2_check2_recon.recon import Recon
-from th2_check2_recon.reconcommon import ReconMessage, _get_msg_timestamp
+from th2_check2_recon.reconcommon import ReconMessage, _get_msg_timestamp, MessageGroupType
 from th2_check2_recon.services import Cache, MessageComparator
 
-
 logger = logging.getLogger(__name__)
 
 
 class Rule:
 
-    def __init__(self, recon: Recon, cache_size: int, match_timeout: int, configuration) -> None:
+    def __init__(self, recon: Recon, cache_size: int, match_timeout: int, autoremove_timeout: Optional[int],
+                 configuration) -> None:
+        self.recon = recon
         self.configure(configuration)
         self.name = self.get_name()
         logger.info("Rule '%s' initializing...", self.name)
 
-        self.recon = recon
         self.event_store = recon.event_store
         self.message_comparator: Optional[MessageComparator] = recon.message_comparator
         self.match_timeout = match_timeout
+        self.autoremove_timeout = autoremove_timeout
+
+        self.reprocess_queue = list()
 
         self.rule_event: Event = \
             EventUtils.create_event(name=self.name,
                                     parent_id=recon.event_store.root_event.id,
-                                    body=EventUtils.create_event_body(MessageComponent(message=self.get_description())))
+                                    body=EventUtils.create_event_body(MessageComponent(message=self.get_description())),
+                                    type=EventUtils.EventType.RULE)
         logger.debug("Created report Event for Rule '%s': %s", self.name,
                      text_format.MessageToString(self.rule_event, as_one_line=True))
         self.event_store.send_parent_event(self.rule_event)
 
-        self.__cache = Cache(self.description_of_groups_bridge(), cache_size, self.event_store, self.rule_event)
+        self.__cache = Cache(self, cache_size)
 
-        self.compared_groups: Dict[str, tuple] = {}  # {ReconMessage.group_id: (Cache.MessageGroup, ..)}
-        for group_id in self.description_of_groups_bridge():
-            self.compared_groups[group_id] = tuple(
-                mg for mg in self.__cache.message_groups if mg.id != group_id)
-
-        self.RULE_PROCESSING_TIME = Histogram(f'th2_recon_{self.name.lower()}_rule_processing_time',
-                                              'Time of processing rule',
-                                              buckets=common_metrics.DEFAULT_BUCKETS)
+        self.RULE_PROCESSING_TIME = Histogram(
+            f"th2_recon_{re.sub('[^a-zA-Z0-9_: ]', '', self.name).lower().replace(' ', '_')}_rule_processing_time",
+            'Time of the message processing with a rule',
+            buckets=common_metrics.DEFAULT_BUCKETS)
 
         logger.info("Rule '%s' initialized", self.name)
 
@@ -86,13 +89,20 @@ class Rule:
         """
         pass
 
+    @property
+    def common_configuration(self):
+        """
+        Returns recon-level rule configuration if it was set, else - None.
+        """
+        return self.recon._config.configuration
+
     def description_of_groups_bridge(self) -> dict:
         result = dict()
-        for (key, value) in self.description_of_groups().items():
-            if type(value) is not set:
-                result[key] = {value}
-            else:
+        for key, value in self.description_of_groups().items():
+            if isinstance(value, set):
                 result[key] = value
+            else:
+                result[key] = {value}
         return result
 
     @abstractmethod
@@ -114,53 +124,85 @@ class Rule:
         return MessageHandler(self)
 
     def put_shared_message(self, shared_group_id: str, message: ReconMessage, attributes: tuple):
+        if shared_group_id in message.in_shared_groups:
+            return
         new_message = copy.deepcopy(message)
         new_message.group_id = shared_group_id
         self.recon.put_shared_message(shared_group_id, new_message, attributes)
 
     def process(self, message: ReconMessage, attributes: tuple, *args, **kwargs):
+        self.clear_out_of_timeout(message.timestamp)
         self.check_no_match_within_timeout(message.timestamp)
 
         self.__group_and_store_event(message, attributes, *args, **kwargs)
         if message.group_id is None:
-            logger.debug("RULE '%s': Ignored  %s", self.name, message.get_all_info())
+            logger.debug("RULE '%s': Ignored %s due to unset group_id", self.name, message.all_info)
             return
 
         self.__hash_and_store_event(message, attributes, *args, **kwargs)
+        if message.hash is None:
+            logger.debug("RULE '%s': Ignored %s due to unset hash", self.name, message.all_info)
+            return
 
-        index_of_main_group = self.__cache.index_of_group(message.group_id)
-        if index_of_main_group == -1:
+        if message.group_id not in self.__cache.message_groups:
             raise Exception(F"'group' method set incorrect groups.\n"
-                            F" - message: {message.get_all_info()}\n"
+                            F" - message: {message.all_info}\n"
                             F" - available groups: {self.description_of_groups_bridge()}\n"
                             F" - message.group_id: {message.group_id}")
-        else:
-            logger.debug("RULE '%s': Received %s", self.name, message.get_all_info())
+        main_group = self.__cache.message_groups[message.group_id]
+        match_all = MessageGroupType.multi_match_all in main_group.type
+        logger.debug("RULE '%s': Received %s", self.name, message.all_info)
 
-        group_indices = []
-        group_sizes = []
-
-        # Check if each group has messages with compared hash else put the message to cache
-        for compared_group in self.compared_groups[message.group_id]:
-            if message.hash not in compared_group:
-                self.__cache.message_groups[index_of_main_group].put(message)
+        for match in self.find_matched_messages(message, match_whole_list=match_all):
+            if match is None:  # Will return None if some of the groups did not contain the message.
+                if MessageGroupType.shared in self.description_of_groups_bridge()[message.group_id] and \
+                        message.group_id not in message.in_shared_groups:
+                    message.in_shared_groups.add(message.group_id)
+                main_group.put(message)
                 return
-            group_indices.append(0)
-            group_sizes.append(len(compared_group.get(message.hash)))
+            else:
+                self.__check_and_store_event(match, attributes, *args, **kwargs)
 
-        group_indices[-1] = -1
-        while self.__increment_indices(group_sizes, group_indices):
-            matched_messages = [message]
-            for i in range(len(group_indices)):
-                index_of_compared_group = i if i < index_of_main_group else i + 1
-                matched_messages.append(
-                    self.__cache.message_groups[index_of_compared_group].data[message.hash][group_indices[i]])
-            self.__check_and_store_event(matched_messages, attributes, *args, **kwargs)
+        if match_all:
+            main_group.put(message)
 
-        for group in self.compared_groups[message.group_id]:
-            if group.is_cleanable:
+        for group_id, group in self.__cache.message_groups.items():
+            if group_id != message.group_id and group.is_cleanable:
                 group.remove(message.hash)
 
+        for message in self.reprocess_queue:
+            self.process(message, attributes)
+        self.reprocess_queue.clear()
+
+    def find_matched_messages(self, message, match_whole_list=False):
+
+        if len(self.description_of_groups_bridge()) == 1:
+            yield [message]
+            return
+
+        matched_messages = list()
+        for group_id, group in self.__cache.message_groups.items():
+            if group_id == message.group_id:
+                continue
+            if message.hash not in group:
+                yield None
+            matched_messages.append(group.data[message.hash])
+
+        if match_whole_list:
+            yield [message] + functools.reduce(lambda inner_list1, inner_list2: inner_list1 + inner_list2,
+                                               matched_messages)
+            return
+
+        for match in zip(*matched_messages):
+            yield [message] + list(match)
+
+    def queue_for_retransmitting(self, message: ReconMessage):
+        self.reprocess_queue.append(message)
+
+    def clear_cache(self):
+        for group in self.__cache.message_groups.values():
+            group.wipe()
+
     def __group_and_store_event(self, message: ReconMessage, attributes: tuple, *args, **kwargs):
         try:
             self.group(message, attributes, *args, **kwargs)
@@ -169,7 +211,7 @@ class Rule:
             self.event_store.store_error(rule_event_id=self.rule_event.id,
                                          event_name="An exception was caught while running 'group'",
                                          error_message=traceback.format_exc(),
-                                         messages=message)
+                                         messages=[message])
 
     def __hash_and_store_event(self, message: ReconMessage, attributes: tuple, *args, **kwargs):
         try:
@@ -179,7 +221,7 @@ class Rule:
             self.event_store.store_error(rule_event_id=self.rule_event.id,
                                          event_name="An exception was caught while running 'hash'",
                                          error_message=traceback.format_exc(),
-                                         messages=message)
+                                         messages=[message])
 
     def __check_and_store_event(self, matched_messages: List[ReconMessage], attributes: tuple, *args, **kwargs):
         for msg in matched_messages:
@@ -208,34 +250,29 @@ class Rule:
                                            check_event=check_event)
 
     def log_groups_size(self):
-        res = ""
-        for group in self.__cache.message_groups:
-            res += f"'{group.id}': {group.size} msg, "
-        res = "[" + res.strip(" ,") + "]"
-        return res
+        return "[" + ', '.join(
+            f"'{group.id}': {group.size} msg" for group in self.__cache.message_groups.values()) + "]"
 
     def cache_size(self):
-        res = 0
-        for group in self.__cache.message_groups:
-            res += group.size
-        return res
+        return sum(group.size for group in self.__cache.message_groups)
 
     def check_no_match_within_timeout(self, actual_time: int):
-        for group in self.__cache.message_groups:
+        for group in self.__cache.message_groups.values():
             group.check_no_match_within_timeout(actual_time, self.match_timeout)
 
+    def clear_out_of_timeout(self, actual_timestamp: int):
+        if self.autoremove_timeout is None:
+            return
+        elif isinstance(self.autoremove_timeout, datetime.datetime):
+            autoremove_timestamp = self.autoremove_timeout.timestamp() * 1e9
+            if autoremove_timestamp < actual_timestamp:
+                self.autoremove_timeout += datetime.timedelta(days=1)
+                self.clear_cache()
+            return
+        elif isinstance(self.autoremove_timeout, int):
+            autoremove_timestamp = actual_timestamp - self.autoremove_timeout * 1e9
+        for group in self.__cache.message_groups.values():
+            group.clear_out_of_timeout(autoremove_timestamp)
+
     def stop(self):
         self.__cache.stop()
-
-    @staticmethod
-    def __increment_indices(sizes: list, indices: list) -> bool:
-        indices[-1] += 1
-        for i in range(len(sizes) - 1, -1, -1):
-            if indices[i] == sizes[i]:
-                if i == 0:
-                    return False
-                indices[i] = 0
-                indices[i - 1] += 1
-            else:
-                break
-        return True
